/** 
 *   FIXME ....
 *
 *   Commands definitions for "antenna" position system.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ??? 
 **/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>

#include <curses.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "iar_bus.h"
#include "iar.h"
#include "my_stdio.h"
#include "survey.h"
#include "iar_bus.h"
#include "antenna_ib_pk.h"
#include "coordinate.h"
#include "encoders.h"
#include "anemometer.h"

/* A structure which contains information on the commands this program
   can understand. */
typedef struct cm {
	char *name;		/* User printable name of the function. */
	int show;               /* show or hide command */
	rl_icpfunc_t *func;	/* Function to call to do the job. */
	char *param;            /* Parameters for this function.  */
	char *doc;		/* Documentation for this function.  */
	struct cm *extended;
} COMMAND;


//typedef struct com COMMAND;

/* The names of functions that actually do the manipulation. */
int com_help __P((char *));
int com_quit __P((char *));
int com_getraw __P((char *));
#ifdef SURVEY
int com_start __P((char *));
int com_stop __P((char *));
int com_status __P((char *));
#endif
int com_show __P((char *));
int com_set __P((char *));
int com_set_zero __P((char *));
int com_set_toggle_zero __P((char *));
int com_set_cw __P((char *));
int com_set_ccw __P((char *));
int com_test_sclk __P((char *));
int com_put_fail_clk __P((char *));
int com_put_normal_clk __P((char *));

COMMAND commands_set[] = {
	{ "zero", 1, com_set_zero, "", 
	  "Set zero for current position on encoder \"dec\" or \"ra\"" },
	{ "togglezero", 1, com_set_toggle_zero, "", 
	  "toggle zero relay at current position on encoder \"dec\" or \"ra\"" },
	{ "cw", 1, com_set_cw, "", 
	  "Set clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ "ccw", 1, com_set_ccw, "", 
	  "Set counter clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ (char *)NULL, 0, (rl_icpfunc_t *)NULL, (char *) NULL, (char *)NULL }
};

COMMAND commands[] = {
#if 0
	{ "setzero", 0, com_set_zero, "", 
	  "Set zero for current position on encoder \"dec\" or \"ra\"" },
	{ "setcw", 0, com_set_cw, "", 
	  "Set clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ "setccw", 0, com_set_ccw, "", 
	  "Set counter clockwhise increment on encoder \"dec\" or \"ra\"" },
#endif
	{ "set", 1, com_set, "", 
	  "Set some encoders attributes (see help)", commands_set},
#ifdef SURVEY
	{ "start", 1, com_start, "", "Start acquisition", NULL},
	{ "stop", 1, com_stop, "", "Stop acquisition", NULL },
	{ "status", 1, com_status, "", "Status of survey file", NULL },
#endif
	{ "show", 1, com_show, "", "Show positon continuously (ENTER: stop)"},

	{ "getenc", 1, com_getraw, "", "Get raw position of encoders" },
	{ "testsclk", 1, com_test_sclk, "", "Test if SCLK is running."},
	{ "failclk", 1, com_put_fail_clk, "", 
	  "Put system in fail condition when SCLK not running"},
	{ "normalclk", 1, com_put_normal_clk, "", 
	  "Put system in normal condition when SCLK running again"},
	{ "help", 1, com_help, "", "Display this text"},
	{ "?", 0, com_help, "", "Synonym for `help'"},
	{ "quit", 1, com_quit, "", "Quit ..."},
	{ "q", 0, com_quit, "", "Quit ..."},
	{ (char *)NULL, 0, (rl_icpfunc_t *)NULL, (char *) NULL, (char *)NULL }
};

/* When non-zero, this means the user is done using this program. */
static int done = 0;
static int enc = -1;

static int ra = RA_CASE;
static int dec = DEC_CASE;

static int readline_flag = 1;
static int interactive_flag = 1;


int my_system (char *command) {
           int pid, status;

           if (command == 0)
               return 1;
           pid = fork();
           if (pid == -1)
               return -1;
           if (pid == 0) {
               char *argv[4];
               argv[0] = "sh";
               argv[1] = "-c";
               argv[2] = command;
               argv[3] = 0;
               execve("/bin/sh", argv, environ);
               exit(127);
           }
           do {
               if (waitpid(pid, &status, 0) == -1) {
                   if (errno != EINTR)
                       return -1;
               } else
                   return status;
           } while(1);
}





/* Generator function for command completion.  STATE lets us
   know whether to start from scratch; without any state
   (i.e. STATE == 0), then we start at the top of the list. */
char *command_generator (char *text, int state)
{ 
	static int list_index, len;
	char *name;
	
	/* If this is a new word to complete, initialize now.  This
	   includes saving the length of TEXT for efficiency, and
	   initializing the index variable to 0. */
	if (!state) {
		list_index = 0;
		len = strlen (text);
	}

	/* Return the next name which partially matches from the
	   command list. */
	while ((name = commands[list_index].name)) {
		list_index++;
		if (strncmp (name, text, len) == 0)
			return (strdup(name));
	}

	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}


char **fileman_completion (const char *text, int start, int end)
{
	char **matches;

	matches = (char **)NULL;

	/* If this word is at the start of the line, then it is a
	   command to complete.  Otherwise it is the name of a file in
	   the current directory.  */
	if (start == 0)
	  matches = rl_completion_matches (text, 
					   (rl_compentry_func_t *) command_generator);
	
	return (matches);
}


/* Tell the GNU Readline library how to complete.  We want to try to
   complete on command names if this is the first word in the line, or
   on filenames if not. */
void initialize_readline (void)
{
//	Keymap km;
//	char f9_macro[]={"\\C-p\\n"};
//	char f5_macro[]={"expose\\n"};
//	char f6_macro[]={"pal last\\n"};

	/* Allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = "antenna";
	
	/* Tell the completer that we want a crack first. */
	rl_attempted_completion_function = fileman_completion;

	/* Read keybindings and variable assignments from filename */
	/* int rl_read_init_file (const char *filename); */

//	km = rl_get_keymap_by_name("emacs"); 
//	rl_macro_bind("\e[15~", f5_macro, km);
//	rl_macro_bind("\e[17~", f6_macro, km);
//	rl_macro_bind("\e[20~", f9_macro, km);
}


/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
COMMAND *find_command (COMMAND *cm, char *name)
{
	int i;

	for (i = 0; cm[i].name; i++)
		if (strcmp (name, cm[i].name) == 0)
			return (&cm[i]);

	return ((COMMAND *)NULL);
}


/* Execute a command line. */
int execute_line (COMMAND *cm, char *line)
{
  int i;
	COMMAND *command;
	char *word;
	int pos_null = 0;
	/* Isolate the command word. */
	i = 0;
	while (line[i] && WHITESPACE(line[i]))	i++;
	word = line + i;

	while (line[i] && !WHITESPACE (line[i])) i++;

	if (line[i]){
		pos_null = i;
		line[i++] = '\0';
	}
	command = find_command (cm, word);
	if (!command) {
		if(pos_null)
			line[pos_null] = ' ';
		my_system(line);
		//fprintf (stderr, "%s: No such command for antennad.\n", word);
		return (-1);
	}

	/* Get argument to command, if any. */
	while (WHITESPACE (line[i])) i++;

	word = line + i;

	/* Call the function. */
	return ((*(command->func)) (word));
}



/* The user wishes to quit using this program.  Just set DONE
   non-zero. */
int com_quit (char *arg)
{
	done = 1;
	return 0;
}


/* Print out help for ARG, or for all of the commands if ARG is not
   present. */
int com_help_commands (COMMAND *cm, char *arg, char *prefix)
{
	int i, j;
	int printed = 0;
	int cmdlen;
	const int param_legend = 22;
	char *sub_arg = arg;
	int len, pos = 0;
	int match;

	len = strlen(arg);
	i = 0;
	while (*(sub_arg+i) != ' ' && i < len) {
		i++;
	}
	if (i) {
		*(sub_arg+i) = 0;
		pos = i == len ? i : i + 1;
	}
	for (i = 0; cm[i].name; i++) {
		match = strcmp(arg, cm[i].name);
		if ((!*arg && cm[i].show) || (match == 0) ) {
			cmdlen = strlen(cm[i].name) + 
				strlen(cm[i].param) + 1;
			if (*prefix != 0) {
				printf ("%s ", prefix);
				cmdlen += strlen(prefix) + 1;
			}
			printf ("%s %s", cm[i].name, 
				cm[i].param);
			if (cmdlen >= param_legend) {
				printf("\n");
				cmdlen = 0;
			}
			for(j = cmdlen; j < param_legend; j++)
				printf(" ");
			printf ("%s.\n", cm[i].doc);
			if (cm[i].extended != NULL && match == 0) {
				com_help_commands(cm[i].extended, 
						  sub_arg+pos, cm[i].name);
			}
			printed++;
		}
	}

	if (!printed) {
		printf ("No commands match `%s'.  Possibilties are:\n", arg);

		for (i = 0; cm[i].name; i++) {
			/* Print in six columns. */
			if (printed == 6) {
				printed = 0;
				printf ("\n");
			}
			
			printf ("%s\t", cm[i].name);
			printed++;
		}

		if (printed)
			printf ("\n");
	}
	return (0);
}


int com_help (char *arg)
{
	return com_help_commands (commands, arg, "");
}



#define COMMAND_LINE_LEGEND 80


char *local_readline(char *prompt)
{
        char *line = NULL;
        fd_set rfds;
        struct timeval tv;
        int retval;
        char c;
        int l;

	write(0, prompt, strlen(prompt));
        line = (char *) malloc(COMMAND_LINE_LEGEND); 

        /* Loop reading and executing lines until the user quits. */
        l = 0;
        do {
                FD_ZERO(&rfds);
                FD_SET(0, &rfds);
                tv.tv_sec = 10;
                tv.tv_usec = 0;
                retval = select(50, &rfds, NULL, NULL, &tv);
                c = 0;
                if (retval) {
                        if (FD_ISSET(0, &rfds)) {
                                /* Win telnet clients send char by char ... */
                                if( read(0, &c, 1) != 1) {
                                        free(line);
                                        return NULL;
                                }
                                *(line+l) = c;
                                l++;
                        }
                }
        } while(c != '\n' && l < COMMAND_LINE_LEGEND && !done);
        if ( *(line+l-2) == 0xd ) *(line+l-2) = 0;
        else *(line+l-1) = 0;

        return line;
}


int command_main(void)
{
	char *line, *s;

	if (readline_flag)
		initialize_readline ();	/* Bind our completer. */

	/* Loop reading and executing lines until the user quits. */
	for ( ; done == 0; ) {
		if( readline_flag)
			line = readline ("antenna> ");
		else 
                        line = local_readline ("antenna> ");

		if (!line)
			break;
		s = stripwhite (line);

		if (*s)
		{
			add_history (s);
			execute_line (commands, s);
		}
		
		free (line);
	}
	return (0);
}


int key_in_buff(int fd)
{
	fd_set rfds;
        int fd_max = fd + 1;
        int retval;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 1;
	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	retval = select(fd_max, &rfds, NULL, NULL, &tv);
	if (retval) {
		if( FD_ISSET(fd, &rfds) ) {
			return 1;
		}
	}

	return 0;
}


int com_set(char * name)
{
	if ( *name == '\0') {
		printf("# set: missing argument.\r\n");
		return (-1);
	}

	return execute_line(commands_set, name);
}


#ifdef SURVEY
int com_start(char * name)
{
	return survey_start(enc);
}


int com_stop(char * name)
{
	return survey_stop(enc);
}


int com_status(char * name)
{
	char *msg = survey_status(enc);
	if (msg != NULL) {
		printf("%s\n", msg);
		free(msg);
		return 0;
	}

	return -1;
}
#endif

int com_show(char * name)
{
	struct encoders_data ed;
	int track_pos = 1;
	char key;

	encoders_start(enc);
	while (track_pos) {
		if (encoders_get_next_data(enc, &ed) == 0) {
#if USE_ANEMOMETER
			printf("%5d (%6.4f) %5d (%7.4f) %3d\n", 
			       ed.encoder[ra],  ENC2DEG(ed.val[ra]),
			       ed.encoder[dec], 
			       -ENC2DEG(ed.val[dec]) /* + IAR_LATITUDE */,
			       anemometer_conv(ed.anemometer));
#else
			printf("%5d (%6.4f) %5d (%7.4f)\n", 
			       ed.encoder[ra],  ENC2DEG(ed.val[ra]),
			       ed.encoder[dec], 
			       -ENC2DEG(ed.val[dec]) /* + IAR_LATITUDE */
			       );
#endif
		}
		fflush(stdout);
		if (key_in_buff(1)) {
			read(1, &key, 1);
			if (key == 0x0a) {
				fflush(stdout);
				track_pos = 0;
			}
		}
	}
	encoders_stop(enc);
	/* FIXME: clear bufer !! */

	return 0;
}


int com_set_zero(char *arg)
{	
        switch( str2cord_type(stripwhite(arg)) ){
        case DEC_CASE:
                printf("# set declination encoder to 0.\n");
                return encoder_set_zero(enc, DEC_CASE);
        case RA_CASE:
                printf("# set right accession encoder to 0.\n");
                return encoder_set_zero(enc, RA_CASE);
        default:
                printf("# set zero: argument error: %s\n", arg);
                return -1;
        }
}
int com_set_toggle_zero(char *arg)
{	
        switch( str2cord_type(stripwhite(arg)) ){
        case DEC_CASE:
                printf("# toggle declination zero relay.\n");
                return encoder_set_toggle_zero(enc, DEC_CASE);
        case RA_CASE:
                printf("# set right accession zero raley.\n");
                return encoder_set_toggle_zero(enc, RA_CASE);
        default:
                printf("# set togglezero: argument error: %s\n", arg);
                return -1;
        }
}


int com_set_cw(char *arg)
{
        switch( str2cord_type(stripwhite(arg)) ){
        case DEC_CASE:
                printf("# set dec. encoder to clockwhise increment.\n");
                return encoder_set_cw(enc, DEC_CASE);
        case RA_CASE:
                printf("# set ra encoder to clockwhise increment.\n");
                return encoder_set_cw(enc, RA_CASE);
        default:
                printf("# set cw: argument error: %s\n", arg);
                return -1;
        }
}


int com_set_ccw(char *arg)
{
        switch( str2cord_type(stripwhite(arg)) ){
        case DEC_CASE:
                printf("# set dec encoder to counter clockwhise increment.\n");
                return encoder_set_ccw(enc, DEC_CASE);
        case RA_CASE:
                printf("# set ra encoder to counter clockwhise increment.\n");
                return encoder_set_ccw(enc, RA_CASE);
        default:
                printf("# set ccw: argument error: %s\n", arg);
                return -1;
        }
}


int com_test_sclk(char *arg)
{
        int res;

        if ( (res = sidereal_clock_running(enc)) == 1) {
                printf("# Sidereal Clock running.\n");
        }
        else {
                printf("# Sidereal Clock NOT running.\n");
        }
        printf("r: %d\n", res);

        return res;
}


int com_put_fail_clk(char *arg)
{
        int res;

        switch ( (res = put_on_fail_condition(enc)) ){
        case 1:
                printf("# Is current condition\n");
                printf("r: 2\n");
                break;
        case 0:
                printf("# Now in fail condition\n");
                printf("r: 1\n");
                break;
        default:
                printf("# ERROR: continuing in normal condition\n");
                printf("r: 0\n");
        }

        return res;
}


int com_put_normal_clk(char *arg)
{
	int res;

        switch ( (res = put_on_normal_condition(enc)) ){
        case 1:
                printf("# Is current condition\n");
                printf("r: 2\n");
                break;
        case 0:
                printf("# Now in normal condition\n");
                printf("r: 1\n");
                break;
        default:
                printf("# ERROR: continuing in fail condition\n");
                printf("r: -1\n");
        }

        return res;
}


int com_getraw(char *arg)
{
        printf("%5d %5d\n", get_encoder(enc, 0), get_encoder(enc, 1) );

        return 0;

}



/*****************************************************************************
 * main:
 */

static void cartel(void)
{
	static const char *msg[] =
	{
	"Antenna Position System ",
	"",
	"Version: ""VERSION",
	"http://www.iar.unlp.edu.ar/~fede/pub/posicionador",
	"",
	"(c) Copyright 2004, 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,",
	"Instituto Argentino de Radio Astronomia (IAR).",
	"",
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License as",
	"published by the Free Software Foundation; either version 2 of",
	"the License, or (at your option) any later version.",
	"",     
	"The author does NOT admit liability nor provide warranty for any",
	"of this software. This material is provided \"AS-IS\" in the hope",
	"that it may be useful for others.",
	""
	};
	unsigned j;
	for (j = 0; j < sizeof(msg)/sizeof(msg[0]); j++)
		fprintf(stderr, "\t%s\n", msg[j]);
	
    return;
}


int main(int argc, char * const argv[])
{
	int op;

	verbose_flag = PERR;
	while  ((op = getopt(argc, argv, "vn")) != EOF)
		switch( op ){
		case 'n':
                        readline_flag = 0;
			interactive_flag = 0;
                        break;
		case 'v':
			verbose_flag = (verbose_flag << 1) | 1;
			break;
		default:
			printf("Use %s -[v (incremental)][n]\n", argv[0]);
			exit(-1);
		}
	if (interactive_flag)
		cartel();

	enc = encoders_open(ra, dec);

	if( enc > 0 ) {
		command_main();
		encoders_close(enc);
	}else{
		PVERB(PERR," antenna: can not connect to encoders\n\n");
	}

        return 0;
}


/*****************************************************************************/
